import os
import sys
import shutil
import stat
import time
import hashlib
from clvm_tools_rs import compile_clvm

TESTDIR = 'testdir'
COMPILED_HASH = '6956b7c3af37e220819dc6e4919e7a1058167192db8eec81714218c4c3d8286d'
SOURCE_FILE = './resources/tests/did_innerpuz.clsp'
TARGET_FILE = './testdir/did_innerpuz.clvm.hex'
SEARCH_PATHS = ['./resources/tests']

def hash_file(path):
    hasher = hashlib.sha256()
    hasher.update(open(path).read().encode('utf8'))
    return hasher.hexdigest()

def test_compile_clvm(source, output, paths, want_sha256):
    compile_clvm(source, output, paths)
    if want_sha256 is not None:
        have_digest = hash_file(output)
        assert want_sha256 == have_digest

def update_file(fname):
    source_content = open(fname).read()
    time.sleep(5.0)
    # Rewrite file.
    with open(fname,'w') as f:
        f.write(source_content)

def try_delete_dir(path):
    try:
        os.chmod(path, stat.S_IREAD | stat.S_IWRITE | stat.S_IXUSR)
        shutil.rmtree(path)
    except:
        pass

def lock_file(testdir, path):
    os.chmod(path, stat.S_IREAD)
    os.chmod(testdir, stat.S_IREAD | stat.S_IXUSR)

def unlock_file(testdir, path):
    os.chmod(testdir, stat.S_IREAD | stat.S_IWRITE | stat.S_IXUSR)
    os.chmod(path, stat.S_IREAD | stat.S_IWRITE)

try_delete_dir(TESTDIR)
os.mkdir(TESTDIR)

# Get baseline date
source_file_date = os.stat(SOURCE_FILE).st_mtime

# Verify that we recompile when the target file isn't present.
test_compile_clvm(SOURCE_FILE, TARGET_FILE, SEARCH_PATHS, COMPILED_HASH)

hex_mod_date = os.stat(TARGET_FILE).st_mtime
assert hex_mod_date > source_file_date

# Update the source file.
update_file(SOURCE_FILE)
source_file_updated_date = os.stat(SOURCE_FILE).st_mtime
assert source_file_updated_date > hex_mod_date
assert source_file_updated_date > source_file_date

# Ensure that we updated the date.
test_compile_clvm(SOURCE_FILE, TARGET_FILE, SEARCH_PATHS, COMPILED_HASH)

updated_hex_mod_date = os.stat(TARGET_FILE).st_mtime
assert updated_hex_mod_date > hex_mod_date
assert updated_hex_mod_date > source_file_updated_date

time.sleep(5.0)
# Mess with the target file, ensure it gets recompiled.
with open(TARGET_FILE,'a') as f:
    f.write('0a0a0a0a0a')

update_file(SOURCE_FILE)
test_compile_clvm(SOURCE_FILE, TARGET_FILE, SEARCH_PATHS, COMPILED_HASH)
# Coming out of here means the new hash matched.
second_updated_hex_mod_date = os.stat(TARGET_FILE).st_mtime

# Update the hex file again and protect it.
update_file(TARGET_FILE)
lock_file(TESTDIR, TARGET_FILE)
hex_date_when_protected = os.stat(TARGET_FILE).st_mtime

update_file(SOURCE_FILE)
test_compile_clvm(SOURCE_FILE, TARGET_FILE, SEARCH_PATHS, COMPILED_HASH)
third_updated_hex_mod_date = os.stat(TARGET_FILE).st_mtime
assert third_updated_hex_mod_date == hex_date_when_protected

# Unprotect and modify.  Compilation should throw because the content won't be
# what's expected.
unlock_file(TESTDIR, TARGET_FILE)
time.sleep(5.0)
# Mess with the target file, ensure it gets recompiled.
with open(TARGET_FILE,'a') as f:
    f.write('0a0a0a0a0a')
lock_file(TESTDIR, TARGET_FILE)

# Update the time on the source file so we'll try to recompile
update_file(SOURCE_FILE)

# Check that we got the wrong hash because the file was modified.
try:
    test_compile_clvm(SOURCE_FILE, TARGET_FILE, SEARCH_PATHS, None)
    have_digest = hash_file(TARGET_FILE)
    assert False
except:
    # We threw because the file wasn't right but we couldn't overwrite it.
    assert COMPILED_HASH != hash_file(TARGET_FILE)

# Set it writable and recompile.
unlock_file(TESTDIR, TARGET_FILE)
test_compile_clvm(SOURCE_FILE, TARGET_FILE, SEARCH_PATHS, COMPILED_HASH)

last_source_mtime = os.stat(SOURCE_FILE).st_mtime
last_hex_mtime = os.stat(TARGET_FILE).st_mtime
assert last_hex_mtime > last_source_mtime



import os
import json
import shutil
import subprocess
from pathlib import Path
import distill_coverage_results

env = dict(os.environ)
env['RUSTFLAGS'] = "-C instrument-coverage"
env['LLVM_PROFILE_FILE'] = "your_name-%p-%m.profraw"

coverage_file = 'coverage.json'

def get_profile_files(the_dir):
    return list(filter(lambda x: x.endswith('.profraw'), os.listdir(the_dir)))

def delete_coverage_files():
    for the_file in get_profile_files('.'):
        if the_file.endswith('.profraw'):
            os.unlink(the_file)

    try:
        os.unlink(coverage_file)
    except:
        pass

def run_coverage_test():
    subprocess.check_call(['cargo','test'],env=env)

def is_my_code(desc):
    for path in ['.cargo','library/std','src/py','src/classic/bins']:
        if path in desc['name']:
            return False

    return True

def collect_coverage():
    profile_files = get_profile_files('.')
    args = ['grcov','--binary-path','target/debug','--output-type','covdir','--output-path',coverage_file] + profile_files
    subprocess.check_call(args)
    result = distill_coverage_results.CoverageResult()

    infile = json.loads(open(coverage_file).read())
    for root in infile['children'].keys():
        distill_coverage_results.distill_coverage(result, Path(root), infile['children'][root])
    return list(filter(is_my_code, result.show_result()))

if __name__ == '__main__':
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument('--require-percent', help="required percentage to succeed", default=None)
    args = parser.parse_args()

    delete_coverage_files()
    run_coverage_test()
    result = collect_coverage()
    print(json.dumps(result, indent=4))

    if args.require_percent is not None:
        required_percentage = int(args.require_percent)
        has_required_pct = True

        for file in result:
            have_pct = int(file['coveragePercent'])
            if have_pct < required_percentage:
                print(f"{file['name']} lacks required coverage have {have_pct} want {required_percentage}")
                has_required_pct = False

        if not has_required_pct:
            sys.exit(1)




import os
from pathlib import Path
import binascii
import json
from clvm_tools.binutils import assemble, disassemble
from clvm_tools_rs import start_clvm_program, compose_run_function, compile_clvm
from chia.types.blockchain_format.program import Program

def compile_module_with_symbols(include_paths,source):
    path_obj = Path(source)
    file_path = path_obj.parent
    file_stem = path_obj.stem
    target_file = file_path / (file_stem + ".clvm.hex")
    sym_file = file_path / (file_stem + ".sym")
    compile_result = compile_clvm(source, str(target_file.absolute()), include_paths, True)
    symbols = compile_result['symbols']
    if len(symbols) != 0:
        with open(str(sym_file.absolute()),'w') as symfile:
            symfile.write(json.dumps(symbols))

def run_until_end(p):
    last = None
    location = None

    while not p.is_ended():
        step_result = p.step()
        if step_result is not None:
            last = step_result
            if 'Result-Location' in last and \
               'Arguments' in last and \
               'print' in last['Result-Location'] and \
               'Value' in last:
                value_assembled = assemble(last['Value'])
                assembled_arg_list = []
                while value_assembled.pair is not None:
                    assembled_arg_list.append(value_assembled.pair[0])
                    value_assembled = value_assembled.pair[1]
                if len(assembled_arg_list) > 2:
                    to_show = disassemble(Program.to(assembled_arg_list[-2:]))
                    print(f"print {last['Operator']} {to_show}")

    return last

def diag_run_clvm(program, args, symbols):
    hex_form_of_program = binascii.hexlify(bytes(program)).decode('utf8')
    hex_form_of_args = binascii.hexlify(bytes(args)).decode('utf8')
    symbols = json.loads(open(symbols).read())
    p = start_clvm_program(hex_form_of_program, hex_form_of_args, symbols)
    report = run_until_end(p)
    if 'Failure' in report:
        print(report['Failure'])

if __name__ == '__main__':
    # smoke test
    import sys
    program = Program.fromhex(open(sys.argv[1]).read())
    args = Program.fromhex(open(sys.argv[2]).read())
    diag_run_clvm(program, args)




from ir import reader, writer


def do_test(sexp_text):
    ir_sexp = reader.read_ir(sexp_text)
    sexp_text_normalized = writer.write_ir(ir_sexp)
    ir_sexp_2 = reader.read_ir(sexp_text)
    sexp_text_normalized_2 = writer.write_ir(ir_sexp_2)
    assert sexp_text_normalized == sexp_text_normalized_2


def test_writer_1():
    do_test("100")

    do_test("0x0100")

    do_test("0x100")

    do_test('"100"')

    do_test('"the quick brown fox jumps over the lazy dogs"')

    do_test("(the quick brown fox jumps over the lazy dogs)")

    do_test("foo")

    do_test("(100 0x0100)")

    do_test('(c (quote 100) (c (quote "foo") (quote ())))')

    do_test("(c . foo)")

    do_test("(a b c de f g h i . j)")



from getopt import GetoptError, getopt
from typing import Dict, List

_options = [
    "exec-prefix=",
    "home=",
    "install-base=",
    "install-data=",
    "install-headers=",
    "install-lib=",
    "install-platlib=",
    "install-purelib=",
    "install-scripts=",
    "prefix=",
    "root=",
    "user",
]


def parse_distutils_args(args: List[str]) -> Dict[str, str]:
    """Parse provided arguments, returning an object that has the matched arguments.

    Any unknown arguments are ignored.
    """
    result = {}
    for arg in args:
        try:
            parsed_opt, _ = getopt(args=[arg], shortopts="", longopts=_options)
        except GetoptError:
            # We don't care about any other options, which here may be
            # considered unrecognized since our option list is not
            # exhaustive.
            continue

        if not parsed_opt:
            continue

        option = parsed_opt[0]
        name_from_parsed = option[0][2:].replace("-", "_")
        value_from_parsed = option[1] or "true"
        result[name_from_parsed] = value_from_parsed

    return result



#!/usr/bin/env python
from testbuild import compile_programs

programs = [
    ('gameref21/rockpaperscissorsd.clsp',['./gameref21']),
    ('gameref21/rockpaperscissorsc.clsp',['./gameref21']),
    ('gameref21/rockpaperscissorsb.clsp',['./gameref21']),
    ('gameref21/rockpaperscissorsa.clsp',['./gameref21']),
    ('gameref21/referee_accuse.clsp',['./gameref21']),
    ('gameref21/referee.clsp',['./gameref21']),
    ('bridgeref/merkle_tree_a2c.clsp',['./bridge-includes']),
    ('bridgeref/validation_taproot.clsp',['./bridge-includes'])
]

compile_programs(programs)



import hashlib

from hsms.streamables.program import Program
from hsms.puzzles.load_clvm import load_clvm

from clvm.EvalError import EvalError


MOD_A = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsa.clvm.hex").read()))
MOD_B = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsb.clvm.hex").read()))
MOD_C = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsc.clvm.hex").read()))
MOD_D = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsd.clvm.hex").read()))


def drun(prog: Program, *args: Program):
    try:
        return prog.run(*args)
    except EvalError as ee:
        print(f"brun -x -y main.sym {prog} {Program.to(list(args))}")
        raise

def sha256(blob:bytes) -> bytes:
    return hashlib.sha256(blob).digest()

def testrps(amove, bmove):
    total = 100
    alice_final = (total//2 if amove == bmove else (0 if bmove == (amove + 1) % 3 else total))
    alice_preimage = Program.to(60 + amove)
    bob_preimage = Program.to(60 + bmove)
    alice_image = sha256(alice_preimage.atom)
    bob_image = sha256(bob_preimage.atom)
    alice_move = Program.to(amove)

    cd = MOD_D.curry(alice_move, bob_image)
    assert cd.run([total, bob_preimage, b'', alice_final, b'j']).atom == b'j'
    cc = MOD_C.curry(alice_image, bob_image)
    assert cc.run([total, alice_preimage, cd.tree_hash(), 0, b'j']).atom == b'j'
    cb = MOD_B.curry(alice_image)
    assert cb.run([total, bob_image, cc.tree_hash(), 0, b'j']).atom == b'j'
    assert MOD_A.run([total, alice_image, cb.tree_hash(), 0, b'j']).atom == b'j'

def testall():
    for i in range(3):
        for j in range(3):
            testrps(i, j)

if __name__ == '__main__':
    testall()



from ir.reader import read_ir
from ir.writer import write_ir


def test_tokenize_comments():
    script_source = "(equal 7 (+ 5 ;foo bar\n   2))"
    expected_output = "(equal 7 (+ 5 2))"
    t = read_ir(script_source)
    s = write_ir(t)
    assert s == expected_output



import sys

from clvm_tools_rs.clvm_tools_rs import cmds


# this dynamic bit of programming may look super-complicated,
# but it's just me being lazy and automating with a macro to
# produce code of the form:
#  brun = lambda: cmds.brun_main(sys.argv)
#  run = lambda: cmds.run_main(sys.argv)
# etc.


for cmd in "brun run opc opd cldb shrink clisp_to_json repl".split():
    code = f"{cmd} = lambda: cmds.{cmd}_main(sys.argv)"
    exec(code)



from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Sequence

if TYPE_CHECKING:
    from pip._vendor.rich.console import ConsoleRenderable

from . import get_console
from .segment import Segment
from .terminal_theme import DEFAULT_TERMINAL_THEME

if TYPE_CHECKING:
    from pip._vendor.rich.console import ConsoleRenderable

JUPYTER_HTML_FORMAT = """\
<pre style="white-space:pre;overflow-x:auto;line-height:normal;font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace">{code}</pre>
"""


class JupyterRenderable:
    """A shim to write html to Jupyter notebook."""

    def __init__(self, html: str, text: str) -> None:
        self.html = html
        self.text = text

    def _repr_mimebundle_(
        self, include: Sequence[str], exclude: Sequence[str], **kwargs: Any
    ) -> Dict[str, str]:
        data = {"text/plain": self.text, "text/html": self.html}
        if include:
            data = {k: v for (k, v) in data.items() if k in include}
        if exclude:
            data = {k: v for (k, v) in data.items() if k not in exclude}
        return data


class JupyterMixin:
    """Add to an Rich renderable to make it render in Jupyter notebook."""

    __slots__ = ()

    def _repr_mimebundle_(
        self: "ConsoleRenderable",
        include: Sequence[str],
        exclude: Sequence[str],
        **kwargs: Any,
    ) -> Dict[str, str]:
        console = get_console()
        segments = list(console.render(self, console.options))
        html = _render_segments(segments)
        text = console._render_buffer(segments)
        data = {"text/plain": text, "text/html": html}
        if include:
            data = {k: v for (k, v) in data.items() if k in include}
        if exclude:
            data = {k: v for (k, v) in data.items() if k not in exclude}
        return data


def _render_segments(segments: Iterable[Segment]) -> str:
    def escape(text: str) -> str:
        """Escape html."""
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    fragments: List[str] = []
    append_fragment = fragments.append
    theme = DEFAULT_TERMINAL_THEME
    for text, style, control in Segment.simplify(segments):
        if control:
            continue
        text = escape(text)
        if style:
            rule = style.get_html_style(theme)
            text = f'<span style="{rule}">{text}</span>' if rule else text
            if style.link:
                text = f'<a href="{style.link}" target="_blank">{text}</a>'
        append_fragment(text)

    code = "".join(fragments)
    html = JUPYTER_HTML_FORMAT.format(code=code)

    return html


def display(segments: Iterable[Segment], text: str) -> None:
    """Render segments to Jupyter."""
    html = _render_segments(segments)
    jupyter_renderable = JupyterRenderable(html, text)
    try:
        from IPython.display import display as ipython_display

        ipython_display(jupyter_renderable)
    except ModuleNotFoundError:
        # Handle the case where the Console has force_jupyter=True,
        # but IPython is not installed.
        pass


def print(*args: Any, **kwargs: Any) -> None:
    """Proxy for Console print."""
    console = get_console()
    return console.print(*args, **kwargs)



import json
from pathlib import Path

class CoverageResult:
    def __init__(self):
        self.info = {}

    def add_file(self,path,name,obj):
        target_path = path / name
        self.info[str(target_path)] = obj

    def show_result(self):
        results = []
        for filename in self.info.keys():
            file_result_dict = {"name":filename}
            for key in ['coveragePercent','linesCovered','linesMissed']:
                file_result_dict[key] = self.info[filename][key]

            results.append(file_result_dict)

        return results

def distill_coverage(result,path,container):
    if not 'children' in container:
        nice_cov_dict = {}
        for item in ['coveragePercent','linesCovered','linesMissed']:
            nice_cov_dict[item] = container[item]
        result.add_file(path, path, nice_cov_dict)
        return

    for subpath in container['children'].keys():
        distill_coverage(result, path / subpath, container['children'][subpath])

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('covfile')
    args = parser.parse_args()

    result = CoverageResult()
    infile = json.loads(open(args.covfile).read())
    for root in infile['children'].keys():
        distill_coverage(result, Path(root), infile['children'][root])

    print(json.dumps(result.show_result(), indent=4))



import pytest
from hashlib import sha256
from contextlib import asynccontextmanager
from chia.clvm.spend_sim import SimClient, SpendSim
from pathlib import Path
from clvm.casts import int_to_bytes, int_from_bytes

from hsms.streamables.program import Program
from clvm_tools_rs import compile_clvm
from clvm_tools.binutils import disassemble

from clvm.EvalError import EvalError
from chia.types.mempool_inclusion_status import MempoolInclusionStatus
from chia.util.errors import Err
from dataclasses import dataclass
from typing import Any
from chia_rs import Coin
from chia.types.spend_bundle import SpendBundle
from chia.types.coin_spend import CoinSpend
from blspy import G2Element

from steprun import diag_run_clvm, compile_module_with_symbols

compile_module_with_symbols(['.'],'referee.clsp')
referee = Program.from_bytes(bytes.fromhex(open("referee.clvm.hex").read()))
refhash = referee.tree_hash()
compile_module_with_symbols(['.'],'referee_accuse.clsp')
referee_accuse = Program.from_bytes(bytes.fromhex(open("referee_accuse.clvm.hex").read()))
refaccusehash = referee.tree_hash()
compile_clvm('rockpaperscissorsa.clsp', 'rockpaperscissorsa.clvm.hex', ['.'])
MOD_A = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsa.clvm.hex").read()))
compile_clvm('rockpaperscissorsb.clsp', 'rockpaperscissorsb.clvm.hex', ['.'])
MOD_B = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsb.clvm.hex").read()))
compile_clvm('rockpaperscissorsc.clsp', 'rockpaperscissorsc.clvm.hex', ['.'])
MOD_C = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsc.clvm.hex").read()))
compile_clvm('rockpaperscissorsd.clsp', 'rockpaperscissorsd.clvm.hex', ['.'])
MOD_D = Program.from_bytes(bytes.fromhex(open("rockpaperscissorsd.clvm.hex").read()))

move = 0
accuse = 1
timeout = 2

def drun(prog: Program, *args: Program):
    try:
        return prog.run(*args)
    except EvalError as ee:
        print(f"brun -x -y main.sym {prog} {Program.to(list(args))}")
        raise

def sha(blob:bytes) -> bytes:
    return sha256(blob).digest()

@pytest.fixture(scope="function")
@asynccontextmanager
async def setup_sim() :
    sim = await SpendSim.create(db_path=Path("file:db_test?mode=memory&cache=shared"))
    sim_client = SimClient(sim)
    await sim.farm_block()

    try:
        yield sim, sim_client
    finally:
        await sim.close()

def bootstrap_referee(parent_coin_id, initial_validation_program_hash, initial_split, 
        amount, timeout, max_move_size, mover_puzzle, waiter_puzzle):
    """
    returns referee_wrap
    """
    puzzle_hash = referee.curry(
        [initial_validation_program_hash, 0, initial_split, amount, timeout, max_move_size, mover_puzzle.tree_hash(), 
        waiter_puzzle.tree_hash(), referee.tree_hash()]).tree_hash()
    coin = Coin(parent_coin_id, puzzle_hash, amount)
    return RefereeWrap(coin, bytes(32), bytes(32), bytes(32),
            initial_validation_program_hash, 0, initial_split, timeout, max_move_size, 
            mover_puzzle, waiter_puzzle)

@dataclass
class RefereeWrap:
    coin: Any
    grandparent_id: Any
    parent_validation_program_hash: Any
    parent_everything_else_hash: Any
    validation_program_hash: Any
    move: Any
    split: Any
    timeout: Any
    max_move_size: Any
    mover_puzzle: Any
    waiter_puzzle: Any

    def curried_parameters_for_our_puzzle(self, purpose, for_self, move_to_make, split, validation_program_hash):
        result = Program.to([
            validation_program_hash,
            move_to_make,
            split,
            self.coin.amount,
            self.timeout,
            self.max_move_size,
            self.mover_puzzle.tree_hash() if for_self else self.waiter_puzzle.tree_hash(),
            self.waiter_puzzle.tree_hash() if for_self else self.mover_puzzle.tree_hash(),
            refhash
        ])
        print(f'for {purpose} curried_parameters_for_our_puzzle is {result}')
        return result

    def get_puzzle(self):
        return referee.curry(self.curried_parameters_for_our_puzzle(
            "GET_PUZZLE",
            True,
            self.move,
            self.split,
            self.validation_program_hash
        ))

    def SpendMove(self, password, move_to_make, split, validation_program_hash):
        """
        returns (solution, new RefereeWrap)
        """
        print(f"MOVE referee mover_puzzle {self.mover_puzzle.tree_hash()}")
        print(f"MOVE referee waiter_puzzle {self.waiter_puzzle.tree_hash()}")
        curried_parameters = self.curried_parameters_for_our_puzzle(
            "SPEND_MOVE",
            False,
            move_to_make,
            split,
            validation_program_hash
        )
        print(f"MOVE referee curried parameters {curried_parameters}")
        new_puzzle_hash = referee.curry(curried_parameters).tree_hash()
        print(f"MOVE new puzzle hash {Program.to(new_puzzle_hash)}")
        solution = Program.to([move, move_to_make, split, validation_program_hash, self.mover_puzzle, 
                               [password, [51, new_puzzle_hash, self.coin.amount]]])
        coin = Coin(self.coin.name(), new_puzzle_hash, self.coin.amount)
        everything_else_hash = Program.to([self.move, self.split, self.coin.amount, self.timeout, 
                self.max_move_size, self.mover_puzzle.tree_hash(), self.waiter_puzzle.tree_hash(), 
                referee.tree_hash()]).tree_hash()
        return (solution, RefereeWrap(coin, self.coin.parent_coin_info, self.validation_program_hash, everything_else_hash,
            validation_program_hash, move_to_make, split, self.timeout, self.max_move_size,
            self.waiter_puzzle, self.mover_puzzle))

    def SpendAccuse(self, password):
        """
        returns (solution, RefereeAccuse)
        """
        print(f"ACCUSE starting with puzzle hash {Program.to(self.get_puzzle().tree_hash())}")
        print(f"ACCUSE parent_id {Program.to(self.coin.parent_coin_info)}")
        print(f"ACCUSE referee mover_puzzle {self.mover_puzzle.tree_hash()}")
        print(f"ACCUSE referee waiter_puzzle {self.waiter_puzzle.tree_hash()}")
        new_puzzle_hash = referee_accuse.curry([
            self.parent_validation_program_hash,
            self.validation_program_hash,
            self.move,
            self.split,
            self.coin.amount,
            self.timeout,
            self.waiter_puzzle.tree_hash(),
            self.mover_puzzle.tree_hash()
        ]).tree_hash()
        solution = Program.to([accuse, self.grandparent_id, self.parent_validation_program_hash,
                self.parent_everything_else_hash, self.mover_puzzle, [password, [51, new_puzzle_hash, self.coin.amount]]])
        coin = Coin(self.coin.name(), new_puzzle_hash, self.coin.amount)
        return (solution, RefereeAccuseWrap(coin, self.parent_validation_program_hash, self.validation_program_hash,
                self.move, self.split, self.timeout, self.waiter_puzzle.tree_hash(),
                self.mover_puzzle.tree_hash()))

    def SpendTimeout(self):
        """
        returns (solution, movercoinid, waitercoinid)
        """
        movercoinid = Coin(self.coin.name(), self.mover_puzzle.tree_hash(), self.split).name()
        waitercoinid = Coin(self.coin.name(), self.waiter_puzzle.tree_hash(), 
                self.coin.amount - self.split).name()
        return (Program.to((timeout, 0)), movercoinid, waitercoinid)

@dataclass
class RefereeAccuseWrap:
    coin: Any
    old_validation_puzzle_hash: Any
    new_validation_puzzle_hash: Any
    move: Any
    split: Any
    timeout: Any
    accused_puzzle_hash: Any
    accuser_puzzle_hash: Any

    def get_puzzle(self):
        return referee_accuse.curry([self.old_validation_puzzle_hash, self.new_validation_puzzle_hash,
                self.move, self.split, self.coin.amount, self.timeout, self.accused_puzzle_hash,
                self.accuser_puzzle_hash])

    def SpendTimeout(self):
        """
        returns (solution, coinid)
        """
        coin = Coin(self.coin.name(), self.accuser_puzzle_hash, self.coin.amount)
        return (Program.to(0), coin.name())

    def SpendDefend(self, validation_program_reveal, validation_program_solution):
        """
        returns (solution, coinid)
        """
        solution = Program.to([validation_program_reveal, validation_program_solution])
        coin = Coin(self.coin.name(), self.accused_puzzle_hash, self.coin.amount)
        return (solution, coin.name())

@pytest.mark.asyncio
@pytest.mark.parametrize('amove', [0, 1, 2])
@pytest.mark.parametrize('bmove', [0, 1, 2])
async def test_rps(amove, bmove, setup_sim):
    total = 100
    alice_final = (total//2 if amove == bmove else (0 if bmove == (amove + 1) % 3 else total))
    alice_preimage = int_to_bytes(60 + amove)
    alice_image = sha(alice_preimage)
    bob_preimage = int_to_bytes(60 + bmove)
    bob_image = sha(bob_preimage)
    alice_move = int_to_bytes(amove)
    nil = Program.to(0)

    # (mod (password . conditions) (if (= password 'alice') conditions (x)))
    alice_puzzle = Program.from_bytes(bytes.fromhex('ff02ffff03ffff09ff02ffff0185616c69636580ffff0103ffff01ff088080ff0180'))
    alice_puzzle_hash = alice_puzzle.tree_hash()
    # (mod (password . conditions) (if (= password 'bob') conditions (x)))
    bob_puzzle = Program.from_bytes(bytes.fromhex('ff02ffff03ffff09ff02ffff0183626f6280ffff0103ffff01ff088080ff0180'))
    bob_puzzle_hash = bob_puzzle.tree_hash()

    async with setup_sim as (sym, client):
        acs = Program.to(1)
        acs_hash = acs.tree_hash()
        await sym.farm_block(acs_hash)
        mycoin = (await client.get_coin_records_by_puzzle_hashes([acs_hash], include_spent_coins = False))[0].coin
        # make a coin for a game
        referee = bootstrap_referee(mycoin.name(), MOD_A.tree_hash(), 2, total, 1000, 50, alice_puzzle, bob_puzzle)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(mycoin, acs, Program.to([[51, referee.coin.puzzle_hash,
                referee.coin.amount]]))], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        savepoint = sym.block_height
        # Alice accuse Bob of cheating (negative test, should fail)
        solution, accuse = referee.SpendAccuse('alice')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(referee.coin, referee.get_puzzle(),
                solution)], G2Element()))
        print(solution)
        print(accuse.get_puzzle())
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.ASSERT_MY_PARENT_ID_FAILED
        # timeout too early fail
        solution, alice_reward_id, bob_reward_id = referee.SpendTimeout()
        spend = SpendBundle([CoinSpend(referee.coin, referee.get_puzzle(), solution)], G2Element())
        (status, err) = await client.push_tx(spend)
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.ASSERT_SECONDS_RELATIVE_FAILED
        # timeout succeeds
        sym.pass_time(2000)
        await sym.farm_block()
        (status, err) = await client.push_tx(spend)
        assert status == MempoolInclusionStatus.SUCCESS
        assert err is None
        await sym.farm_block()
        assert (await client.get_coin_records_by_names([alice_reward_id], include_spent_coins = False))[0].coin.amount == 2
        assert (await client.get_coin_records_by_names([bob_reward_id], include_spent_coins = False))[0].coin.amount == total - 2
        await sym.rewind(savepoint)
        # Alice makes an illegally large move, fails
        solution, ref2 = referee.SpendMove('alice', bytes(100), 0, bytes(32))
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(referee.coin, referee.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        # Alice makes move with negative split, fails
        solution, ref2 = referee.SpendMove('alice', 'abc', -1, bytes(32))
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(referee.coin, referee.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        # Alice makes move with split greater than amount, fails
        solution, ref2 = referee.SpendMove('alice', 'abc', referee.coin.amount + 1, bytes(32))
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(referee.coin, referee.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        # Alice move 1 commit to image
        bpuz = MOD_B.curry(alice_image)
        solution, ref2 = referee.SpendMove('alice', alice_image, 0, bpuz.tree_hash())
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(referee.coin, referee.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        savepoint = sym.block_height
        # Bob accuse Alice of cheating
        solution, accuse = ref2.SpendAccuse('bob')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref2.coin, ref2.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        savepoint2 = sym.block_height
        # Alice accusation defend, gets everything
        solution, reward_id = accuse.SpendDefend(MOD_A, nil)
        print(solution)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        reward_coin_wrapper = await client.get_coin_records_by_names([reward_id], include_spent_coins = 
                False)
        reward_coin = reward_coin_wrapper[0].coin
        assert reward_coin.amount == referee.coin.amount
        assert reward_coin.puzzle_hash == alice_puzzle_hash
        await sym.rewind(savepoint2)
        # accusation timeout too early fail
        solution, reward_id = accuse.SpendTimeout()
        spend = SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), solution)], G2Element())
        (status, err) = await client.push_tx(spend)
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.ASSERT_SECONDS_RELATIVE_FAILED
        # accusation timeout succeed, Bob gets everything
        sym.pass_time(2000)
        await sym.farm_block()
        (status, err) = await client.push_tx(spend)
        assert status == MempoolInclusionStatus.SUCCESS
        assert err is None
        await sym.farm_block()
        reward_coin_wrapper = await client.get_coin_records_by_names([reward_id], include_spent_coins = 
                False)
        reward_coin = reward_coin_wrapper[0].coin
        assert reward_coin.amount == referee.coin.amount
        assert reward_coin.puzzle_hash == bob_puzzle_hash
        await sym.rewind(savepoint)
        # Bob move 2 commit to image
        cpuz = MOD_C.curry([alice_image, bob_image])
        solution, ref3 = ref2.SpendMove('bob', bob_image, 0, cpuz.tree_hash())
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref2.coin, ref2.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        savepoint = sym.block_height
        # Alice accuse
        solution, accuse = ref3.SpendAccuse('alice')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref3.coin, ref3.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Bob defends
        solution, reward_id = accuse.SpendDefend(bpuz, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        reward_coin = (await client.get_coin_records_by_names([reward_id], include_spent_coins = 
                False))[0].coin
        assert reward_coin.amount == referee.coin.amount
        assert reward_coin.puzzle_hash == bob_puzzle_hash
        await sym.rewind(savepoint)
        # Alice reveals wrong preimage
        alice_bad_preimage = int_to_bytes(61 + amove)
        dpuz = MOD_D.curry([(amove + 1) % 3, bob_image])
        solution, ref4 = ref3.SpendMove('alice', alice_bad_preimage, 0, dpuz.tree_hash())
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref3.coin, ref3.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Bob accuses
        solution, accuse = ref4.SpendAccuse('bob')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref4.coin, ref4.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Alice defends, fails
        solution, reward_id = accuse.SpendDefend(cpuz, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        await sym.rewind(savepoint)
        # Alice move 3 reveal preimage
        dpuz = MOD_D.curry([alice_move, bob_image])
        solution, ref4 = ref3.SpendMove('alice', alice_preimage, 0, dpuz.tree_hash())
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref3.coin, ref3.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        savepoint = sym.block_height
        # Bob accuses
        solution, accuse = ref4.SpendAccuse('bob')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref4.coin, ref4.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Alice defends
        solution, reward_id = accuse.SpendDefend(cpuz, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.rewind(savepoint)
        # Bob move 4 reveal wrong preimage
        bob_bad_preimage = int_to_bytes(121 + amove)
        solution, ref5 = ref4.SpendMove('bob', bob_bad_preimage, 0, dpuz.tree_hash())
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref4.coin, ref4.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Alice accuses
        solution, accuse = ref5.SpendAccuse('alice')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref5.coin, ref5.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Bob attempts defense, fails
        solution, reward_id = accuse.SpendDefend(dpuz, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        # Bob attempts defense with wrong validation program, fails
        solution, reward_id = accuse.SpendDefend(acs, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        await sym.rewind(savepoint)
        if amove == bmove:
            # Bob move 4 gives wrong split
            solution, ref5 = ref4.SpendMove('bob', bob_preimage, 0, dpuz.tree_hash())
            (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref4.coin, ref4.get_puzzle(), 
                    solution)], G2Element()))
            assert status == MempoolInclusionStatus.SUCCESS
            await sym.farm_block()
            # Alice accuses
            solution, accuse = ref5.SpendAccuse('alice')
            (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref5.coin, ref5.get_puzzle(), 
                    solution)], G2Element()))
            assert status == MempoolInclusionStatus.SUCCESS
            await sym.farm_block()
            # Bob attempts defense, fails
            solution, reward_id = accuse.SpendDefend(dpuz, nil)
            (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                    solution)], G2Element()))
            assert status == MempoolInclusionStatus.FAILED
            assert err == Err.GENERATOR_RUNTIME_ERROR
            await sym.rewind(savepoint)
        # Bob move 4 reveal preimage
        solution, ref5 = ref4.SpendMove('bob', bob_preimage, alice_final, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref4.coin, ref4.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        savepoint = sym.block_height
        # Alice attempts move, fails
        solution, ref6 = ref5.SpendMove('alice', nil, 0, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref5.coin, ref5.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.FAILED
        assert err == Err.GENERATOR_RUNTIME_ERROR
        # timeout, split correct
        sym.pass_time(2000)
        await sym.farm_block()
        solution, alice_reward_id, bob_reward_id = ref5.SpendTimeout()
        spend = SpendBundle([CoinSpend(ref5.coin, ref5.get_puzzle(), solution)], G2Element())
        (status, err) = await client.push_tx(spend)
        assert status == MempoolInclusionStatus.SUCCESS
        assert err is None
        await sym.farm_block()
        if alice_final != 0:
            assert (await client.get_coin_records_by_names([alice_reward_id], include_spent_coins = False))[0].coin.amount == alice_final
        else:
            assert len(await client.get_coin_records_by_names([alice_reward_id], include_spent_coins = False)) == 0
        if alice_final != ref5.coin.amount:
            assert (await client.get_coin_records_by_names([bob_reward_id], include_spent_coins = False))[0].coin.amount == ref5.coin.amount - alice_final
        else:
            assert len(await client.get_coin_records_by_names([bob_reward_id], include_spent_coins = False)) == 0
        await sym.rewind(savepoint)
        # Alice accuses
        solution, accuse = ref5.SpendAccuse('alice')
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(ref5.coin, ref5.get_puzzle(), 
                solution)], G2Element()))
        assert status == MempoolInclusionStatus.SUCCESS
        await sym.farm_block()
        # Bob defends
        solution, reward_id = accuse.SpendDefend(dpuz, nil)
        (status, err) = await client.push_tx(SpendBundle([CoinSpend(accuse.coin, accuse.get_puzzle(), 
                solution)], G2Element()))
        assert (status, err) == (MempoolInclusionStatus.SUCCESS, None)



from clvm.operators import KEYWORD_TO_ATOM

from clvm_tools.binutils import assemble, disassemble
from clvm_tools.curry import curry, uncurry


def check_idempotency(f, args):
    cost, curried = curry(f, args)

    r = disassemble(curried)
    f_0, args_0 = uncurry(curried)

    assert disassemble(f_0) == disassemble(f)
    assert disassemble(args_0) == disassemble(args)
    return r


def test_curry_uncurry():
    PLUS = KEYWORD_TO_ATOM["+"][0]
    f = assemble("(+ 2 5)")
    args = assemble("(200 30)")
    actual_disassembly = check_idempotency(f, args)
    assert actual_disassembly == f"(a (q {PLUS} 2 5) (c (q . 200) (c (q . 30) 1)))"

    f = assemble("(+ 2 5)")
    args = assemble("((+ (q . 50) (q . 60)))")
    actual_disassembly = check_idempotency(f, args)
    assert actual_disassembly == f"(a (q {PLUS} 2 5) (c (q {PLUS} (q . 50) (q . 60)) 1))"



import os
import subprocess
import traceback

recompile_list = [
    'block_program_zero.clsp',
    'calculate_synthetic_public_key.clsp',
    'chialisp_deserialisation.clsp',
    'decompress_coin_spend_entry.clsp',
    'decompress_coin_spend_entry_with_prefix.clsp',
    'decompress_puzzle.clsp',
    'delegated_tail.clsp',
    'did_innerpuz.clsp',
    'everything_with_signature.clsp',
    'genesis_by_coin_id.clsp',
    'genesis_by_puzzle_hash.clsp',
    'nft_metadata_updater_default.clsp',
    'nft_metadata_updater_updateable.clsp',
    'nft_ownership_layer.clsp',
    'nft_ownership_transfer_program_one_way_claim_with_royalties.clsp',
    'nft_state_layer.clsp',
    'p2_conditions.clsp',
    'p2_delegated_conditions.clsp',
    'p2_delegated_puzzle.clsp',
    'p2_delegated_puzzle_or_hidden_puzzle.clsp',
    'p2_m_of_n_delegate_direct.clsp',
    'p2_puzzle_hash.clsp',
    'p2_singleton.clsp',
    'p2_singleton_or_delayed_puzhash.clsp',
    'pool_member_innerpuz.clsp',
    'pool_waitingroom_innerpuz.clsp',
    'rom_bootstrap_generator.clsp',
    'settlement_payments.clsp',
    'sha256tree_module.clsp',
    'singleton_launcher.clsp',
    'singleton_top_layer.clsp',
    'singleton_top_layer_v1_1.clsp',
    'test_generator_deserialize.clsp',
    'test_multiple_generator_input_arguments.clsp'
]

for fname in recompile_list:
    hexfile = f'./chia/wallet/puzzles/{fname}.hex'
    hexdata = open(hexfile).read().strip()
    os.unlink(hexfile)
    try:
        compiled = subprocess.check_output(['../target/release/run', '-i', 'chia/wallet/puzzles/', f'chia/wallet/puzzles/{fname}']).strip()
        recompile = subprocess.check_output(['../target/release/opc', compiled]).decode('utf8').strip()
    except:
        print(f'compiling {fname}')
        traceback.print_exc()

    if hexdata != recompile:
        print(f'*** COMPILE RESULTED IN DIFFERENT OUTPUT FOR FILE {fname}')
        assert hexdata == recompile



import os
from pathlib import Path
import binascii
import json
from clvm_tools.binutils import assemble, disassemble
from clvm_tools_rs import start_clvm_program, compose_run_function, compile_clvm
from chia_rs import run_chia_program, tree_hash

def compute_output_file_names(source):
    path_obj = Path(source)
    file_path = path_obj.parent
    file_stem = path_obj.stem
    return (file_path / (file_stem + ".clvm.hex"), file_path / (file_stem + ".sym"))

def get_program_hash(hexfile):
    return tree_hash(binascii.unhexlify(open(hexfile).read().strip()))

def compile_module_with_symbols(include_paths,source):
    (hex_file, sym_file) = compute_output_file_names(source)
    compile_result = compile_clvm(source, str(hex_file.absolute()), include_paths, True)
    symbols = compile_result['symbols']
    if len(symbols) != 0:
        with open(str(sym_file.absolute()),'w') as symfile:
            symfile.write(json.dumps(symbols))

def compile_programs(programs):
    for p in programs:
        hex_file, sym_file = compute_output_file_names(p[0])
        prev_hex_file = str(hex_file.absolute()) + '.reference'
        prev_sym_file = str(sym_file.absolute()) + '.reference'
        prev_symbols = json.loads(open(prev_sym_file).read())
        prev_hash = get_program_hash(prev_hex_file)

        try:
            os.unlink(hex_file)
        except:
            pass

        compile_module_with_symbols(p[1], p[0])
        curr_hash = get_program_hash(hex_file)
        curr_symbols = json.loads(open(sym_file).read())

        if prev_hash != curr_hash:
            print(f'compiling {p[0]} got a different hash')
            assert prev_hash == curr_hash

        # Things might be added, but we should produce the same basic
        # outline.
        for hash_match, data in prev_symbols.items():
            # Be lenient about line number info from generated sources.
            if '*' in data:
                continue

            assert hash_match in curr_symbols
            assert curr_symbols[hash_match] == data
